@@ -2,6 +2,7 @@
 This module contains all of the mount/update
 functions for re-wx's supported widget types.
 """
+from operator import attrgetter
 import wx
 import wx.adv
 import wx.media
@@ -10,11 +11,11 @@ import wx.html2
 import wx.svg
 # TODO: warn on unknown props?
 import os
-from wx.lib.scrolledpanel import ScrolledPanel
+import weakref
 from rewx.dispatch import mount, update
 from wx.richtext import RichTextCtrl
 import sys
-from rewx.components import Block, Grid, TextArea, SVG, SVGButton, NotebookItem
+from rewx.components import Block, Grid, FlexGrid, TextArea, SVG, SVGButton, NotebookItem, FilePickerCtrlOpen, FilePickerCtrlSave, DirPickerCtrl, ScrolledPanel
 from rewx.util import exclude
 from rewx.bitmap_support import load, resize_image, to_bitmap
 from rewx.util import identity
@@ -41,6 +42,7 @@ basic_controls = {
     'show': 'Show',
     'enabled': 'Enable',
     'style': 'SetStyle',
+    'wx_name': 'SetName'
 }
 
 exclusions = {
@@ -55,7 +57,14 @@ exclusions = {
     Block: {'style'},
     SVG: {'value'},
     SVGButton: {'value'},
-    ScrolledPanel: {'value'}
+    ScrolledPanel: {'value'},
+    DirPickerCtrl: {'value', 'style'},
+    FilePickerCtrlOpen: {'value', 'style'},
+    FilePickerCtrlSave: {'value', 'style'},
+    FlexGrid: {'value', 'label', 'style'},
+    wx.StaticBitmap: {'value'},
+    wx.ToggleButton: {'value'},
+    wx.ComboBox: {'style'},
 }
 
 
@@ -66,11 +75,10 @@ def set_basic_props(instance, props):
     for key, val in props.items():
         if key.startswith('on_'):
             continue
-        try:
-            getattr(instance, available_controls[key])(val)
-        except KeyError:
-            # prop which doesn't apply to this control
-            pass
+        if key in available_controls:
+            setter_name = available_controls[key]
+            if hasattr(instance, setter_name):
+                getattr(instance, setter_name)(val)
     return instance
 
 
@@ -93,11 +101,19 @@ def frame(element, instance: wx.Frame):
     if 'show' in props:
         instance.Show(props['show'])
     if 'icon_uri' in props:
-        instance.SetIcon(wx.Icon(props['icon_uri']))
-    if 'on_close' in props:
-        instance.Bind(wx.EVT_CLOSE, props['on_close'])
+        icon = wx.Icon(props['icon_uri'])
+        instance.SetIcon(icon)
+        if sys.platform != 'win32':
+            # OSX needs to have its taskbar icon explicitly set
+            # bizarrely, wx requires the TaskBarIcon to be attached to the Frame
+            # as instance data (self.). Otherwise, it will not render correctly.
+            frame.taskbarIcon = wx.adv.TaskBarIcon(iconType=wx.adv.TBI_DOCK)
+            frame.taskbarIcon.SetIcon(icon)
     else:
         instance.SetIcon(wx.Icon(os.path.join(dirname, 'icon.png')))
+    if 'on_close' in props:
+        instance.Bind(wx.EVT_CLOSE, props['on_close'])
+
     return instance
 
 
@@ -117,21 +133,25 @@ def activity_indicator(element, instance: wx.ActivityIndicator):
 
 @mount.register(wx.Button)
 def button(element, parent):
-    default_style = 0
-    return update(element, wx.Button(parent, element['props'].get('style', default_style)))
+    return update(element, wx.Button(parent, style=element['props'].get('style', 0)))
 
 
 @update.register(wx.Button)
 def button(element, instance: wx.Button):
     props = element['props']
     set_basic_props(instance, props)
-    if props.get('enabled') is False:
-        # we navigate away from the control when disabling.
-        # Without this, under conditions where the button gets
-        # disabled / re-enabled quickly, wx will still feed events
-        # spawned while it was disable to the now recently enabled
-        # button. This behavior doesn't happen if we navigate away
-        instance.Navigate()
+    # Removed this behavior because I don't want it, and it causes a problem
+    # because the Button grabs focus on every render when disabled.
+    #
+    # if props.get('enabled') is False:
+    #     # we navigate away from the control when disabling.
+    #     # Without this, under conditions where the button gets
+    #     # disabled / re-enabled quickly, wx will still feed events
+    #     # spawned while it was disable to the now recently enabled
+    #     # button. This behavior doesn't happen if we navigate away
+    #     instance.Navigate()
+    if 'label_markup' in props:
+        instance.SetLabelMarkup(props['label_markup'])
     instance.Unbind(wx.EVT_BUTTON)
     if props.get('on_click'):
         instance.Bind(wx.EVT_BUTTON, props['on_click'])
@@ -233,7 +253,11 @@ def collapsiblepanel(element, instance: wx.CollapsiblePane):
 
 @mount.register(wx.ComboBox)
 def combobox(element, parent):
-    return update(element, wx.ComboBox(parent, choices=element['props'].get('choices', [])))
+    props = element['props']
+    return update(
+        element,
+        wx.ComboBox(parent, choices=props.get('choices', []), style=props.get('style', 0))
+        )
 
 @update.register(wx.ComboBox)
 def combobox(element, instance: wx.ComboBox) -> wx.Object:
@@ -252,7 +276,7 @@ def combobox(element, instance: wx.ComboBox) -> wx.Object:
     for _ in instance.GetItems():
         instance.Delete(0)
     instance.AppendItems(props.get('choices', []))
-    if 'value' in props:
+    if 'value' in element['props']:
         instance.SetSelection(props['choices'].index(element['props'].get('value')))
 
     if props.get('on_change'):
@@ -262,6 +286,98 @@ def combobox(element, instance: wx.ComboBox) -> wx.Object:
 
     return instance
 
+class DirPickerDropTarget(wx.FileDropTarget):
+    """
+    We are required to inherit from wx.FileDropTarget
+    """
+    def __init__(self, dirPickerCtrl, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.dirPickerCtrl = dirPickerCtrl
+    def OnDropFiles(self, x:int, y:int, filenames:list[str]):
+        if len(filenames) == 1:
+            path = filenames[0]
+            if os.path.isdir(path):
+                dpc = self.dirPickerCtrl()
+                if dpc is not None:
+                    if hasattr(dpc, '_on_change'):
+                        dpc._on_change(path)
+                        return True # wx.DragCopy?
+        return False # wx.DragNone?
+    def OnDragOver(self, x:int, y:int, defResult):
+        return wx.DragCopy
+
+@mount.register(DirPickerCtrl)
+def dir_picker_ctrl(element, parent):
+    instance = DirPickerCtrl(parent)
+    # TODO We only need this droptarget in Microsoft Windows.
+    # In Linux this is handled automatically.
+    # https://docs.wxpython.org/wx.FileDropTarget.html
+    instance.droptarget = DirPickerDropTarget(weakref.ref(instance))
+    instance.SetDropTarget(instance.droptarget)
+    return update(element, instance)
+
+@update.register(DirPickerCtrl)
+def dir_picker_ctrl(element, instance: DirPickerCtrl):
+    props = element['props']
+    set_basic_props(instance, props)
+    if instance.HasTextCtrl():
+        textctrl = instance.GetTextCtrl()
+        if 'text_font' in props:
+            textctrl.SetFont(props['text_font'])
+    additions = {
+        'path': 'SetPath'
+    }
+    for prop_key, wx_method in additions.items():
+        if prop_key in props:
+            getattr(instance, wx_method)(props[prop_key])
+    instance.Unbind(wx.EVT_DIRPICKER_CHANGED)
+    if 'on_change' in props:
+        instance._on_change = props['on_change']
+        instance.Bind(wx.EVT_DIRPICKER_CHANGED, instance._on_change_impl)
+    else:
+        if hasattr(instance, '_on_change'):
+            delattr(instance, '_on_change')
+    return instance
+
+@mount.register(FilePickerCtrlOpen)
+def file_picker_ctrl_open(element, parent):
+    return update(element, FilePickerCtrlOpen(parent))
+
+@update.register(FilePickerCtrlOpen)
+def file_picker_ctrl_open(element, instance: FilePickerCtrlOpen):
+    props = element['props']
+    set_basic_props(instance, props)
+    additions = {
+        'path': 'SetPath'
+    }
+    for prop_key, wx_method in additions.items():
+        if prop_key in props:
+            getattr(instance, wx_method)(props[prop_key])
+    instance.Unbind(wx.EVT_FILEPICKER_CHANGED)
+    if props.get('on_change'):
+        instance.Bind(wx.EVT_FILEPICKER_CHANGED, props['on_change'])
+    return instance
+
+@mount.register(FilePickerCtrlSave)
+def file_picker_ctrl_save(element, parent):
+    props = element['props']
+    # TODO This default wildcard prop value is wrong, it should be maybe FileSelectorDefaultWildcardStr?
+    return update(element, FilePickerCtrlSave(parent, style=wx.FLP_SAVE, wildcard=props.get('wildcard', '*.*')))
+
+@update.register(FilePickerCtrlSave)
+def file_picker_ctrl_save(element, instance: FilePickerCtrlSave):
+    props = element['props']
+    set_basic_props(instance, props)
+    additions = {
+        'path': 'SetPath'
+    }
+    for prop_key, wx_method in additions.items():
+        if prop_key in props:
+            getattr(instance, wx_method)(props[prop_key])
+    instance.Unbind(wx.EVT_FILEPICKER_CHANGED)
+    if props.get('on_change'):
+        instance.Bind(wx.EVT_FILEPICKER_CHANGED, props['on_change'])
+    return instance
 
 @mount.register(wx.Gauge)
 def gauge(element, parent):
@@ -321,8 +437,8 @@ def listbox(element, instance: wx.ListBox):
             instance.Delete(0)
         instance.AppendItems(props.get('choices', []))
 
-    for selection in props.get('selected', []):
-        instance.SetSelection(selection)
+    if 'value' in props:
+        instance.SetSelection(element['props'].get('value'))
 
     # TODO: control this component similar to Notebook
     if props.get('on_change'):
@@ -334,7 +450,9 @@ def listbox(element, instance: wx.ListBox):
 
 @mount.register(wx.ListCtrl)
 def listctrl(element, parent):
-    return update(element, wx.ListCtrl(parent, style=wx.LC_REPORT))
+    instance = wx.ListCtrl(parent, style=wx.LC_REPORT)
+    instance.self_managed = True
+    return update(element, instance)
 
 @update.register(wx.ListCtrl)
 def listctrl(element, instance: wx.ListCtrl):
@@ -343,10 +461,12 @@ def listctrl(element, instance: wx.ListCtrl):
         del props['style']
     set_basic_props(instance, props)
     # TODO: what events...?
+    restored_widths = ListCtrl_GetColumnWidths(instance)
     instance.DeleteAllColumns()
     instance.DeleteAllItems()
     for e, col in enumerate(props.get('column_defs', [])):
         instance.InsertColumn(e, col['title'])
+    ListCtrl_SetColumnWidths(instance, restored_widths)
 
     for row_idx, item in enumerate(props.get('data', [])):
         instance.InsertItem(row_idx, '')
@@ -354,6 +474,19 @@ def listctrl(element, instance: wx.ListCtrl):
             instance.SetItem(row_idx, col_idx, coldef['column'](item))
     return instance
 
+def ListCtrl_GetColumnWidths(instance: wx.ListCtrl):
+    wds = []
+    for i in range(0, instance.GetColumnCount()):
+        wds.append(instance.GetColumnWidth(i))
+    return wds
+
+def ListCtrl_SetColumnWidths(instance: wx.ListCtrl, wds: list[int]):
+    # no-op if the list of column widths is the wrong length
+    if len(wds) == instance.GetColumnCount():
+        for i in range(0, len(wds)):
+            instance.SetColumnWidth(i, wds[i])
+
+
 
 @mount.register(wx.media.MediaCtrl)
 def mediactrl(element, parent):
@@ -473,7 +606,6 @@ def staticbox(element, instance: wx.StaticBox):
 
     return instance
 
-
 @mount.register(wx.Slider)
 def slider(element, parent):
     return update(element, wx.Slider(parent))
@@ -502,13 +634,15 @@ def spinctrl(element, instance: wx.SpinCtrl):
     instance.SetMin(props.get('min', 0))
     instance.Unbind(wx.EVT_SPINCTRL)
     if 'on_change' in props:
-        instance.Bind(wx.EVT_SLIDER, props['on_change'])
+        instance.Bind(wx.EVT_SPINCTRL, props['on_change'])
     return instance
 
 
 @mount.register(wx.SpinCtrlDouble)
 def spinctrldouble(element, parent):
-    return update(element, wx.SpinCtrlDouble(parent))
+    instance = wx.SpinCtrlDouble(parent)
+    instance.self_managed = True # SpinCtrlDouble has a self-managed child.
+    return update(element, instance)
 
 @update.register(wx.SpinCtrlDouble)
 def spinctrldouble(element, instance: wx.SpinCtrlDouble):
@@ -519,9 +653,9 @@ def spinctrldouble(element, instance: wx.SpinCtrlDouble):
     instance.SetIncrement(props.get('increment', 0))
     if 'digits' in props:
         instance.SetDigits(props['digits'])
-    instance.Unbind(wx.EVT_SPINCTRL)
+    instance.Unbind(wx.EVT_SPINCTRLDOUBLE)
     if 'on_change' in props:
-        instance.Bind(wx.EVT_SLIDER, props['on_change'])
+        instance.Bind(wx.EVT_SPINCTRLDOUBLE, props['on_change'])
     return instance
 
 
@@ -675,6 +809,9 @@ def block(element, instance: Block):
     instance.Unbind(wx.EVT_LEFT_DOWN)
     if 'on_click' in props:
         instance.Bind(wx.EVT_LEFT_DOWN, props['on_click'])
+    instance.Unbind(wx.EVT_SIZE)
+    if 'on_size' in props:
+        instance.Bind(wx.EVT_SIZE, props['on_size'])
     return instance
 
 
@@ -690,13 +827,47 @@ def scrolledpanel(element, parent):
 def scrolledpanel(element, instance: ScrolledPanel):
     props = element['props']
     set_basic_props(instance, props)
-    instance.SetupScrolling(
-        scroll_x=props.get('scroll_x', False),
-        scroll_y=props.get('scroll_y', False)
-    )
+    # Only call SetupScrolling if the props changed.
+    # https://docs.wxpython.org/wx.lib.scrolledpanel.ScrolledPanel.html#wx.lib.scrolledpanel.ScrolledPanel.SetupScrolling
+    # https://github.com/wxWidgets/Phoenix/blob/de0a4415c05e7483b2960c6dc9720154269244a8/wx/lib/scrolledpanel.py#L120
+    scroll_x = props.get('scroll_x', False)
+    scroll_y = props.get('scroll_y', False)
+    rate_x = props.get('rate_x', 20)
+    rate_y = props.get('rate_y', 20)
+    if (not hasattr(instance, 'last_scroll_x')
+        or instance.last_scroll_x != scroll_x
+        or not hasattr(instance, 'last_scroll_y')
+        or instance.last_scroll_y != scroll_y
+        or not hasattr(instance, 'last_rate_x')
+        or instance.last_rate_x != rate_x
+        or not hasattr(instance, 'last_rate_y')
+        or instance.last_rate_y != rate_y
+        ):
+        instance.SetupScrolling(
+            scroll_x=scroll_x,
+            scroll_y=scroll_y,
+            rate_x=rate_x,
+            rate_y=rate_y
+        )
+    instance.last_scroll_x = scroll_x
+    instance.last_scroll_y = scroll_y
+    instance.last_rate_x = rate_x
+    instance.last_rate_y = rate_y
     instance.Unbind(wx.EVT_LEFT_DOWN)
     if 'on_click' in props:
         instance.Bind(wx.EVT_LEFT_DOWN, props['on_click'])
+    instance.Unbind(wx.EVT_SIZE)
+    if 'on_size' in props:
+        instance.Bind(wx.EVT_SIZE, props['on_size'])
+    instance.Unbind(wx.EVT_MOTION)
+    if 'on_mouse_motion' in props:
+        instance.Bind(wx.EVT_MOTION, props['on_mouse_motion'])
+    instance.Unbind(wx.EVT_MOUSEWHEEL)
+    if 'on_mouse_wheel' in props:
+        instance.Bind(wx.EVT_MOUSEWHEEL, props['on_mouse_wheel'])
+    instance.Unbind(wx.EVT_SCROLLWIN)
+    if 'on_scrollwin' in props:
+        instance.Bind(wx.EVT_SCROLLWIN, props['on_scrollwin'])
     return instance
 
 
@@ -709,7 +880,6 @@ def grid(element, parent):
     panel.SetSizer(sizer)
     return update(element, panel)
 
-
 @update.register(Grid)
 def grid(element, instance: Grid):
     props = element['props']
@@ -720,6 +890,24 @@ def grid(element, instance: Grid):
     return instance
 
 
+@mount.register(FlexGrid)
+def flexgrid(element, parent):
+    props = element['props']
+    panel = FlexGrid(parent)
+    sizer = wx.FlexGridSizer(props.get('cols', 1), gap=props.get('gap', (0, 0)))
+    panel.SetSizer(sizer)
+    return update(element, panel)
+
+@update.register(FlexGrid)
+def flexgrid(element, instance: FlexGrid):
+    props = element['props']
+    set_basic_props(instance, props)
+    instance.Unbind(wx.EVT_LEFT_DOWN)
+    if 'on_click' in props:
+        instance.Bind(wx.EVT_LEFT_DOWN, props['on_click'])
+    return instance
+
+
 @mount.register(wx.TextCtrl)
 def textctrl(element, parent):
     style = element['props'].get('style', wx.TE_LEFT)
@@ -746,6 +934,15 @@ def textctrl(element, instance: wx.TextCtrl):
     instance.Unbind(wx.EVT_TEXT)
     if 'on_change' in props:
         instance.Bind(wx.EVT_TEXT, props['on_change'])
+    instance.Unbind(wx.EVT_SET_FOCUS)
+    if 'on_focus_set' in props:
+        instance.Bind(wx.EVT_SET_FOCUS, props['on_focus_set'])
+    instance.Unbind(wx.EVT_KILL_FOCUS)
+    if 'on_focus_kill' in props:
+        instance.Bind(wx.EVT_KILL_FOCUS, props['on_focus_kill'])
+    instance.Unbind(wx.EVT_KEY_DOWN)
+    if 'on_key_down' in props:
+        instance.Bind(wx.EVT_KEY_DOWN, props['on_key_down'])
     return instance
 
 
@@ -764,6 +961,7 @@ def staticbitmap(element, parent):
 @update.register(wx.StaticBitmap)
 def staticbitmap(element, instance: wx.StaticBitmap):
     props = element['props']
+    set_basic_props(instance, props)
     if props.get('uri'):
         # only load and update the image if it has changed.
         if instance._rewx_cache.get('uri', 'rewx::nothing') != props['uri']:
@@ -853,11 +1051,10 @@ def notebook(element, parent):
 def notebook(element, instance: wx.Notebook):
     props = element['props']
     set_basic_props(instance, props)
-    handler = notebook_selection(props.get('on_change', identity))
-    if instance._changeHandler != handler:
-        instance.Unbind(wx.EVT_NOTEBOOK_PAGE_CHANGED)
-        instance.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, handler)
-        instance._changeHandler = handler
+    # handler = notebook_selection(props.get('on_change', identity))
+    # if instance._changeHandler != handler:
+    instance.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, props.get('on_change', identity))
+        # instance._changeHandler = handler
     return instance
 
 
@@ -938,7 +1135,7 @@ _supported_props = {
             'SetRange': 'range: int', # sets the maximum value, so just needs an int
             'SetValue': 'value: int'
         },
-        'wx.ListBox': {
+        'wx.Listbox': {
             'style': 'style',
             'EVT_LISTBOX': 'on_change',
             'choices': 'choices', # update will use InsertItems,
